"""Tests for the SQLite-to-PostgreSQL migration tooling."""

from __future__ import annotations
import hashlib
import json
import sqlite3
from pathlib import Path
from typing import Any
import pytest
from orcheo.tooling import postgres_migration as migration


class MockCopy:
    """Mock copy context manager."""

    def __init__(self) -> None:
        self.rows: list[tuple[Any, ...]] = []

    def write_row(self, row: tuple[Any, ...]) -> None:
        self.rows.append(row)

    def __enter__(self) -> MockCopy:
        return self

    def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        return None


class TrackingMockCursor:
    """Mock cursor that tracks execute calls."""

    def __init__(self) -> None:
        self.executemany_calls: list[tuple[str, list[tuple[Any, ...]]]] = []
        self.execute_calls: list[tuple[str, Any | None]] = []
        self.copy_calls: list[tuple[str, list[tuple[Any, ...]]]] = []

    def executemany(self, query: str, rows: list[tuple[Any, ...]]) -> None:
        self.executemany_calls.append((query, list(rows)))

    def execute(self, query: str, params: Any | None = None) -> None:
        self.execute_calls.append((query, params))

    def copy(self, query: str) -> MockCopy:
        copier = MockCopy()
        self.copy_calls.append((query, copier.rows))
        return copier

    def __enter__(self) -> TrackingMockCursor:
        return self

    def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        return None


class TrackingMockConnection:
    """Mock connection using TrackingMockCursor."""

    def __init__(self) -> None:
        self.cursor_instance = TrackingMockCursor()
        self.autocommit = False
        self.commits = 0

    def cursor(self) -> TrackingMockCursor:
        return self.cursor_instance

    def commit(self) -> None:
        self.commits += 1

    def __enter__(self) -> TrackingMockConnection:
        return self

    def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        return None


def _write_manifest(manifest_path: Path, payload: dict[str, Any]) -> None:
    manifest_path.write_text(json.dumps(payload, indent=2, sort_keys=True))


def test_transform_row_parses_json_and_bool(tmp_path: Path) -> None:
    db_path = tmp_path / "sample.sqlite"
    with sqlite3.connect(db_path) as conn:
        conn.execute("CREATE TABLE sample (payload TEXT, is_best INTEGER)")
        conn.execute(
            "INSERT INTO sample (payload, is_best) VALUES (?, ?)",
            (json.dumps({"ok": True}), 1),
        )
        conn.commit()

    with sqlite3.connect(db_path) as conn:
        conn.row_factory = sqlite3.Row
        row = conn.execute("SELECT payload, is_best FROM sample").fetchone()
        assert row is not None

    spec = migration.TableSpec(
        name="sample",
        sqlite_path=db_path,
        sqlite_table="sample",
        postgres_table="sample",
        columns=("payload", "is_best"),
        order_by=(),
        json_columns=("payload",),
        bool_columns=("is_best",),
    )
    transformed = migration._transform_row(row, spec)

    assert transformed["payload"] == {"ok": True}
    assert transformed["is_best"] is True


def test_export_table_writes_batches(tmp_path: Path) -> None:
    db_path = tmp_path / "workflows.sqlite"
    with sqlite3.connect(db_path) as conn:
        conn.execute(
            "CREATE TABLE workflows "
            "(id TEXT, payload TEXT, created_at TEXT, updated_at TEXT)"
        )
        conn.execute(
            "INSERT INTO workflows VALUES (?, ?, ?, ?)",
            ("wf_1", json.dumps({"name": "one"}), "2024-01-01", "2024-01-01"),
        )
        conn.execute(
            "INSERT INTO workflows VALUES (?, ?, ?, ?)",
            ("wf_2", json.dumps({"name": "two"}), "2024-01-02", "2024-01-02"),
        )
        conn.commit()

    spec = migration.TableSpec(
        name="workflows",
        sqlite_path=db_path,
        sqlite_table="workflows",
        postgres_table="workflows",
        columns=("id", "payload", "created_at", "updated_at"),
        order_by=("id",),
        json_columns=("payload",),
    )

    manifest = migration.export_table(spec, tmp_path / "export", batch_size=1)

    assert manifest.row_count == 2
    assert len(manifest.batches) == 2
    for batch in manifest.batches:
        batch_path = tmp_path / "export" / batch.file
        assert batch_path.exists()
        assert batch.checksum


def test_import_manifest_inserts_batches(tmp_path: Path) -> None:
    export_dir = tmp_path / "export"
    export_dir.mkdir()

    batch_path = export_dir / "workflows" / "batch_0001.jsonl"
    batch_path.parent.mkdir(parents=True, exist_ok=True)
    payload = {"id": "wf_1", "payload": {"name": "one"}}
    line = json.dumps(payload).encode("utf-8") + b"\n"
    batch_path.write_bytes(line)
    checksum = hashlib.sha256(line).hexdigest()

    manifest = {
        "version": 1,
        "generated_at": "now",
        "tables": {
            "workflows": {
                "name": "workflows",
                "sqlite_path": "ignored",
                "postgres_table": "workflows",
                "columns": ["id", "payload"],
                "row_count": 1,
                "batches": [
                    {
                        "file": str(batch_path.relative_to(export_dir)),
                        "rows": 1,
                        "checksum": checksum,
                    }
                ],
                "json_columns": ["payload"],
                "bool_columns": [],
                "post_import_sql": [],
            }
        },
    }
    manifest_path = export_dir / "manifest.json"
    _write_manifest(manifest_path, manifest)

    connection = TrackingMockConnection()

    def connect_stub(_: str) -> TrackingMockConnection:
        return connection

    result = migration.import_manifest(
        manifest_path,
        "postgresql://test",
        connection_factory=connect_stub,
    )

    assert result["tables"] == ["workflows"]
    assert result["tables"] == ["workflows"]
    assert connection.cursor_instance.copy_calls
    copy_sql, rows = connection.cursor_instance.copy_calls[0]
    assert "COPY workflows" in copy_sql
    assert rows[0] == ("wf_1", json.dumps({"name": "one"}))


def test_validate_manifest_reports_mismatches(tmp_path: Path) -> None:
    export_dir = tmp_path / "export"
    export_dir.mkdir()
    manifest = {
        "version": 1,
        "generated_at": "now",
        "tables": {
            "workflows": {
                "name": "workflows",
                "sqlite_path": "ignored",
                "postgres_table": "workflows",
                "columns": ["id"],
                "row_count": 2,
                "batches": [],
                "json_columns": [],
                "bool_columns": [],
                "post_import_sql": [],
            }
        },
    }
    manifest_path = export_dir / "manifest.json"
    _write_manifest(manifest_path, manifest)

    class DummyCursor:
        def execute(self, query: str) -> None:
            return None

        def fetchone(self) -> tuple[int]:
            return (1,)

        def __enter__(self) -> DummyCursor:
            return self

        def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
            return None

    class DummyConnection:
        def cursor(self) -> DummyCursor:
            return DummyCursor()

        def __enter__(self) -> DummyConnection:
            return self

        def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
            return None

    def connect_stub(_: str) -> DummyConnection:
        return DummyConnection()

    result = migration.validate_manifest(
        manifest_path,
        "postgresql://test",
        connection_factory=connect_stub,
    )

    assert result["mismatches"][0]["table"] == "workflows"


def test_normalize_json_edge_cases() -> None:
    from orcheo.tooling.postgres_migration import _normalize_json

    assert _normalize_json(None) is None
    assert _normalize_json("") is None  # Empty string logic
    assert _normalize_json("  ") is None
    assert _normalize_json('{"a": 1}') == {"a": 1}
    # No need to test malformed json as it would raise JSONDecodeError
    # which isn't caught there (that's fine)


def test_transform_row_nulls() -> None:
    # Test boolean column handling when None
    spec = migration.TableSpec(
        name="t",
        sqlite_path=Path("."),
        sqlite_table="t",
        postgres_table="t",
        columns=("b",),
        order_by=(),
        bool_columns=("b",),
    )

    row = {"b": None}
    transformed = migration._transform_row(row, spec)  # type: ignore
    assert transformed["b"] is False

    row_true = {"b": 1}
    transformed_true = migration._transform_row(row_true, spec)  # type: ignore
    assert transformed_true["b"] is True


def test_main_cli(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    # We can mock _build_parser or just pass argv to main
    from orcheo.tooling.postgres_migration import main

    # 1. Export command
    repo_db = tmp_path / "repo.sqlite"
    repo_db.touch()
    chatkit_db = tmp_path / "chatkit.sqlite"
    chatkit_db.touch()

    # Mock build_table_specs and export_all to avoid actual DB work
    monkeypatch.setattr(migration, "build_table_specs", lambda **kwargs: [])
    monkeypatch.setattr(migration, "export_all", lambda *args, **kwargs: {})

    exit_code = main(
        [
            "export",
            "--output",
            str(tmp_path),
            "--repository-sqlite",
            str(repo_db),
            "--chatkit-sqlite",
            str(chatkit_db),
        ]
    )
    assert exit_code == 0

    # 2. Import command
    monkeypatch.setattr(migration, "import_manifest", lambda *args, **kwargs: {})
    manifest = tmp_path / "manifest.json"
    manifest.touch()

    exit_code = main(
        ["import", "--input", str(tmp_path), "--postgres-dsn", "postgres://test"]
    )
    assert exit_code == 0

    # 3. Validate command
    monkeypatch.setattr(
        migration, "validate_manifest", lambda *args, **kwargs: {"ok": True}
    )

    exit_code = main(
        ["validate", "--input", str(tmp_path), "--postgres-dsn", "postgres://test"]
    )
    assert exit_code == 0


def test_import_manifest_resume(tmp_path: Path) -> None:
    # Create manifest and import_state.json
    export_dir = tmp_path / "export"
    export_dir.mkdir()

    # batch1 is already done
    batch1_path = export_dir / "t1" / "b1.jsonl"
    batch1_path.parent.mkdir(parents=True, exist_ok=True)
    batch1_path.write_text('{"a":1}')  # Needs proper content for checksum
    import hashlib

    c1 = hashlib.sha256(b'{"a":1}').hexdigest()

    # batch2 is new
    batch2_path = export_dir / "t1" / "b2.jsonl"
    batch2_path.write_text('{"a":2}')
    c2 = hashlib.sha256(b'{"a":2}').hexdigest()

    manifest = {
        "tables": {
            "t1": {
                "postgres_table": "t1",
                "columns": ["a"],
                "batches": [
                    {"file": "t1/b1.jsonl", "checksum": c1},
                    {"file": "t1/b2.jsonl", "checksum": c2},
                ],
            }
        }
    }
    manifest_path = export_dir / "manifest.json"
    manifest_path.write_text(json.dumps(manifest))

    state_path = export_dir / "import_state.json"
    state_path.write_text(json.dumps({"completed_batches": ["t1/b1.jsonl"]}))

    connection = TrackingMockConnection()

    def connect_stub(_: str) -> TrackingMockConnection:
        return connection

    migration.import_manifest(
        manifest_path, "dsn", resume=True, connection_factory=connect_stub
    )

    # Should only copy batch 2
    assert len(connection.cursor_instance.copy_calls) == 1
    # Check that state was updated
    new_state = json.loads(state_path.read_text())
    assert "t1/b2.jsonl" in new_state["completed_batches"]


def test_export_table_validations(tmp_path: Path) -> None:
    spec = migration.TableSpec("n", Path("missing"), "t", "t", (), ())

    with pytest.raises(ValueError, match="positive"):
        migration.export_table(spec, tmp_path, batch_size=0)

    with pytest.raises(FileNotFoundError):
        migration.export_table(spec, tmp_path, batch_size=1)


def test_normalize_json_with_empty_string(tmp_path: Path) -> None:
    """Test _normalize_json with whitespace-only strings."""
    from orcheo.tooling.postgres_migration import _normalize_json

    assert _normalize_json("   \t\n  ") is None
    assert _normalize_json("") is None


def test_process_table_with_checksum_mismatch(tmp_path: Path) -> None:
    """Test that _process_table raises error on checksum mismatch."""
    export_dir = tmp_path / "export"
    export_dir.mkdir()

    batch_path = export_dir / "t1" / "b1.jsonl"
    batch_path.parent.mkdir(parents=True, exist_ok=True)
    batch_path.write_text('{"a":1}')

    table = {
        "postgres_table": "t1",
        "columns": ["a"],
        "batches": [
            {"file": "t1/b1.jsonl", "checksum": "wrong_checksum"},
        ],
    }

    connection = TrackingMockConnection()
    cursor = connection.cursor()

    with pytest.raises(ValueError, match="Checksum mismatch"):
        migration._process_table(
            connection,
            cursor,
            table,
            export_dir,
            set(),
            {"completed_batches": []},
            export_dir / "state.json",
        )


def test_process_table_with_post_import_sql(tmp_path: Path) -> None:
    """Test that _process_table executes post_import_sql."""
    export_dir = tmp_path / "export"
    export_dir.mkdir()

    batch_path = export_dir / "t1" / "b1.jsonl"
    batch_path.parent.mkdir(parents=True, exist_ok=True)
    line = b'{"a":1}'
    batch_path.write_bytes(line)
    checksum = hashlib.sha256(line).hexdigest()

    table = {
        "postgres_table": "t1",
        "columns": ["a"],
        "batches": [
            {"file": "t1/b1.jsonl", "checksum": checksum},
        ],
        "post_import_sql": ["SELECT setval('seq', 1)"],
    }

    connection = TrackingMockConnection()
    cursor = connection.cursor()
    state_path = export_dir / "state.json"

    migration._process_table(
        connection,
        cursor,
        table,
        export_dir,
        set(),
        {"completed_batches": []},
        state_path,
    )

    # Verify post_import_sql was executed
    assert any("setval" in str(call) for call in cursor.execute_calls)


def test_import_manifest_with_no_resume(tmp_path: Path) -> None:
    """Test import_manifest with resume=False ignores existing state."""
    export_dir = tmp_path / "export"
    export_dir.mkdir()

    batch_path = export_dir / "t1" / "b1.jsonl"
    batch_path.parent.mkdir(parents=True, exist_ok=True)
    line = b'{"a":1}'
    batch_path.write_bytes(line)
    checksum = hashlib.sha256(line).hexdigest()

    manifest = {
        "tables": {
            "t1": {
                "postgres_table": "t1",
                "columns": ["a"],
                "batches": [
                    {"file": "t1/b1.jsonl", "checksum": checksum},
                ],
            }
        }
    }
    manifest_path = export_dir / "manifest.json"
    manifest_path.write_text(json.dumps(manifest))

    # Create old state
    state_path = export_dir / "import_state.json"
    state_path.write_text(json.dumps({"completed_batches": ["t1/b1.jsonl"]}))

    connection = TrackingMockConnection()

    def connect_stub(_: str) -> TrackingMockConnection:
        return connection

    # Import without resume should reprocess everything
    migration.import_manifest(
        manifest_path, "dsn", resume=False, connection_factory=connect_stub
    )

    # Should have processed the batch even though it was in old state
    assert len(connection.cursor_instance.copy_calls) >= 1


def test_validate_manifest_with_psycopg_none(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Test validate_manifest raises error when psycopg is None."""
    monkeypatch.setattr(migration, "psycopg", None)

    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text('{"tables": {}}')

    with pytest.raises(RuntimeError, match="psycopg"):
        migration.validate_manifest(manifest_path, "dsn")


def test_import_manifest_with_psycopg_none(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Test import_manifest raises error when psycopg is None."""
    monkeypatch.setattr(migration, "psycopg", None)

    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text('{"tables": {}}')

    with pytest.raises(RuntimeError, match="psycopg"):
        migration.import_manifest(manifest_path, "dsn")


def test_resolve_paths_missing_paths(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test _resolve_paths raises when paths are missing."""
    monkeypatch.setattr(migration, "get_settings", lambda: {})
    monkeypatch.setattr(
        migration,
        "load_auth_settings",
        lambda: type("obj", (), {"service_token_db_path": None})(),
    )

    with pytest.raises(ValueError, match="Repository and ChatKit"):
        migration._resolve_paths(
            repository_sqlite=None,
            chatkit_sqlite=None,
            service_token_sqlite=None,
        )


def test_resolve_paths_with_fallback_token_path(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Test _resolve_paths uses fallback service token path."""
    repo_path = tmp_path / "repo.sqlite"
    repo_path.touch()
    chatkit_path = tmp_path / "chatkit.sqlite"
    chatkit_path.touch()
    fallback_token_path = tmp_path / "service_tokens.sqlite"
    fallback_token_path.touch()

    monkeypatch.setattr(
        migration,
        "get_settings",
        lambda: {
            "REPOSITORY_SQLITE_PATH": str(repo_path),
            "CHATKIT_SQLITE_PATH": str(chatkit_path),
        },
    )
    monkeypatch.setattr(
        migration,
        "load_auth_settings",
        lambda: type("obj", (), {"service_token_db_path": None})(),
    )

    repo, chatkit, token = migration._resolve_paths(
        repository_sqlite=None,
        chatkit_sqlite=None,
        service_token_sqlite=None,
    )

    assert repo == repo_path
    assert chatkit == chatkit_path
    assert token == fallback_token_path


def test_build_table_specs_without_service_tokens(tmp_path: Path) -> None:
    """Test build_table_specs without service token path."""
    repo_path = tmp_path / "repo.sqlite"
    chatkit_path = tmp_path / "chatkit.sqlite"

    specs = migration.build_table_specs(
        repository_sqlite=repo_path,
        chatkit_sqlite=chatkit_path,
        service_token_sqlite=None,
    )

    # Should not include service_tokens tables
    spec_names = [s.name for s in specs]
    assert "service_tokens" not in spec_names
    assert "service_token_audit_log" not in spec_names
    assert "workflows" in spec_names


def test_main_import_without_postgres_dsn(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Test main import command raises error without POSTGRES_DSN."""
    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text('{"tables": {}}')

    monkeypatch.setattr(migration, "get_settings", lambda: {})

    with pytest.raises(SystemExit, match="POSTGRES_DSN"):
        migration.main(["import", "--input", str(tmp_path)])


def test_main_validate_failure(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    """Test main validate command exits on validation failure."""
    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text('{"tables": {}}')

    monkeypatch.setattr(migration, "get_settings", lambda: {"POSTGRES_DSN": "test"})
    monkeypatch.setattr(
        migration,
        "validate_manifest",
        lambda *args, **kwargs: {"ok": False, "mismatches": [{"table": "test"}]},
    )

    with pytest.raises(SystemExit, match="Validation failed"):
        migration.main(["validate", "--input", str(tmp_path)])


def test_main_unknown_command(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test main with unknown command raises error."""
    mock_args = type(
        "args", (), {"command": "unknown", "postgres_dsn": "test", "input": Path(".")}
    )()

    class MockParser:
        def parse_args(self, argv: list[str] | None = None) -> Any:
            return mock_args

    monkeypatch.setattr(migration, "_build_parser", lambda: MockParser())
    monkeypatch.setattr(migration, "get_settings", lambda: {"POSTGRES_DSN": "test"})

    with pytest.raises(SystemExit, match="Unknown command"):
        migration.main([])


def test_now_iso_returns_isoformat() -> None:
    """Test _now_iso returns an ISO format string."""
    result = migration._now_iso()
    assert isinstance(result, str)
    # Should be parseable as ISO format
    from datetime import datetime

    datetime.fromisoformat(result.replace("Z", "+00:00"))


def test_normalize_json_with_non_string_value() -> None:
    """Test _normalize_json returns non-string values as-is."""
    # Test with dict
    result = migration._normalize_json({"key": "value"})
    assert result == {"key": "value"}

    # Test with list
    result = migration._normalize_json([1, 2, 3])
    assert result == [1, 2, 3]

    # Test with int
    result = migration._normalize_json(42)
    assert result == 42


def test_export_all_generates_manifest(tmp_path: Path) -> None:
    """Test export_all exports multiple tables and generates manifest."""
    db_path = tmp_path / "test.sqlite"
    with sqlite3.connect(db_path) as conn:
        conn.execute("CREATE TABLE t1 (id TEXT)")
        conn.execute("CREATE TABLE t2 (id TEXT)")
        conn.execute("INSERT INTO t1 VALUES ('a')")
        conn.execute("INSERT INTO t2 VALUES ('b')")
        conn.commit()

    specs = [
        migration.TableSpec(
            name="t1",
            sqlite_path=db_path,
            sqlite_table="t1",
            postgres_table="t1",
            columns=("id",),
            order_by=("id",),
        ),
        migration.TableSpec(
            name="t2",
            sqlite_path=db_path,
            sqlite_table="t2",
            postgres_table="t2",
            columns=("id",),
            order_by=("id",),
        ),
    ]

    output_dir = tmp_path / "export"
    result = migration.export_all(specs, output_dir, batch_size=1000)

    # Manifest should have version and tables
    assert result["version"] == 1
    assert "generated_at" in result
    assert "t1" in result["tables"]
    assert "t2" in result["tables"]

    # Manifest file should be written
    manifest_path = output_dir / "manifest.json"
    assert manifest_path.exists()
    stored = json.loads(manifest_path.read_text())
    assert stored["version"] == 1


def test_process_table_with_empty_batch(tmp_path: Path) -> None:
    """Test _process_table handles batch with no rows gracefully."""
    export_dir = tmp_path / "export"
    export_dir.mkdir()

    # Create an empty batch file
    batch_path = export_dir / "t1" / "b1.jsonl"
    batch_path.parent.mkdir(parents=True, exist_ok=True)
    batch_path.write_bytes(b"")  # Empty file
    checksum = hashlib.sha256(b"").hexdigest()

    table = {
        "postgres_table": "t1",
        "columns": ["a"],
        "batches": [
            {"file": "t1/b1.jsonl", "checksum": checksum},
        ],
    }

    connection = TrackingMockConnection()
    cursor = connection.cursor()
    state_path = export_dir / "state.json"

    migration._process_table(
        connection,
        cursor,
        table,
        export_dir,
        set(),
        {"completed_batches": []},
        state_path,
    )

    # Should not have any copy calls since rows list is empty
    assert len(cursor.copy_calls) == 0


def test_validate_manifest_no_mismatches(tmp_path: Path) -> None:
    """Test validate_manifest returns ok when counts match."""
    export_dir = tmp_path / "export"
    export_dir.mkdir()
    manifest = {
        "version": 1,
        "generated_at": "now",
        "tables": {
            "workflows": {
                "name": "workflows",
                "sqlite_path": "ignored",
                "postgres_table": "workflows",
                "columns": ["id"],
                "row_count": 5,  # Expecting 5 rows
                "batches": [],
                "json_columns": [],
                "bool_columns": [],
                "post_import_sql": [],
            }
        },
    }
    manifest_path = export_dir / "manifest.json"
    _write_manifest(manifest_path, manifest)

    class DummyCursor:
        def execute(self, query: str) -> None:
            return None

        def fetchone(self) -> tuple[int]:
            return (5,)  # Return 5 rows to match expected

        def __enter__(self) -> DummyCursor:
            return self

        def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
            return None

    class DummyConnection:
        def cursor(self) -> DummyCursor:
            return DummyCursor()

        def __enter__(self) -> DummyConnection:
            return self

        def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
            return None

    def connect_stub(_: str) -> DummyConnection:
        return DummyConnection()

    result = migration.validate_manifest(
        manifest_path,
        "postgresql://test",
        connection_factory=connect_stub,
    )

    assert result["ok"] is True
    assert result["mismatches"] == []


def test_validate_manifest_fetchone_returns_none(tmp_path: Path) -> None:
    """Test validate_manifest handles fetchone returning None."""
    export_dir = tmp_path / "export"
    export_dir.mkdir()
    manifest = {
        "version": 1,
        "generated_at": "now",
        "tables": {
            "workflows": {
                "name": "workflows",
                "postgres_table": "workflows",
                "row_count": 1,
                "batches": [],
            }
        },
    }
    manifest_path = export_dir / "manifest.json"
    _write_manifest(manifest_path, manifest)

    class DummyCursor:
        def execute(self, query: str) -> None:
            return None

        def fetchone(self) -> None:
            return None  # Return None

        def __enter__(self) -> DummyCursor:
            return self

        def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
            return None

    class DummyConnection:
        def cursor(self) -> DummyCursor:
            return DummyCursor()

        def __enter__(self) -> DummyConnection:
            return self

        def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
            return None

    def connect_stub(_: str) -> DummyConnection:
        return DummyConnection()

    result = migration.validate_manifest(
        manifest_path,
        "postgresql://test",
        connection_factory=connect_stub,
    )

    # count should be 0 when row is None, so there's a mismatch
    assert result["ok"] is False
    assert result["mismatches"][0]["actual"] == 0


def test_resolve_paths_no_fallback_token(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Test _resolve_paths when no fallback token path exists."""
    repo_path = tmp_path / "repo.sqlite"
    repo_path.touch()
    chatkit_path = tmp_path / "chatkit.sqlite"
    chatkit_path.touch()
    # Do NOT create the fallback token path

    monkeypatch.setattr(
        migration,
        "get_settings",
        lambda: {
            "REPOSITORY_SQLITE_PATH": str(repo_path),
            "CHATKIT_SQLITE_PATH": str(chatkit_path),
        },
    )
    monkeypatch.setattr(
        migration,
        "load_auth_settings",
        lambda: type("obj", (), {"service_token_db_path": None})(),
    )

    repo, chatkit, token = migration._resolve_paths(
        repository_sqlite=None,
        chatkit_sqlite=None,
        service_token_sqlite=None,
    )

    assert repo == repo_path
    assert chatkit == chatkit_path
    assert token is None  # No fallback should be found


def test_build_table_specs_with_service_tokens(tmp_path: Path) -> None:
    """Test build_table_specs includes service token tables when path is provided."""
    repo_path = tmp_path / "repo.sqlite"
    chatkit_path = tmp_path / "chatkit.sqlite"
    token_path = tmp_path / "tokens.sqlite"

    specs = migration.build_table_specs(
        repository_sqlite=repo_path,
        chatkit_sqlite=chatkit_path,
        service_token_sqlite=token_path,
    )

    spec_names = [s.name for s in specs]
    assert "service_tokens" in spec_names
    assert "service_token_audit_log" in spec_names

    # Check post_import_sql for audit_log
    audit_spec = next(s for s in specs if s.name == "service_token_audit_log")
    assert len(audit_spec.post_import_sql) > 0
    assert "setval" in audit_spec.post_import_sql[0]


def test_import_manifest_uses_psycopg_connect(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Test import_manifest uses psycopg.connect when no factory is provided."""
    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text('{"tables": {}}')

    connection = TrackingMockConnection()

    def mock_connect(*args: Any, **kwargs: Any) -> TrackingMockConnection:
        return connection

    mock_psycopg = type("obj", (), {"connect": mock_connect})()
    monkeypatch.setattr(migration, "psycopg", mock_psycopg)

    # This should now hit line 262
    result = migration.import_manifest(manifest_path, "dsn", connection_factory=None)
    assert result == {"tables": []}


def test_validate_manifest_uses_psycopg_connect(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Test validate_manifest uses psycopg.connect when no factory is provided."""
    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text('{"tables": {}}')

    connection = TrackingMockConnection()

    def mock_connect(*args: Any, **kwargs: Any) -> TrackingMockConnection:
        return connection

    mock_psycopg = type("obj", (), {"connect": mock_connect})()
    monkeypatch.setattr(migration, "psycopg", mock_psycopg)

    # This should now hit line 293
    result = migration.validate_manifest(manifest_path, "dsn", connection_factory=None)
    assert result["ok"] is True



"""Global embedding registrations and identifiers for conversational search."""

from __future__ import annotations
from collections.abc import Callable
from typing import Final
from langchain_openai import OpenAIEmbeddings
from orcheo.nodes.conversational_search.embeddings import (
    register_langchain_embedding,
    register_pinecone_bm25_embedding,
)
from orcheo.nodes.conversational_search.ingestion import resolve_embedding_method


OPENAI_TEXT_EMBEDDING_3_SMALL: Final[str] = "embedding:openai:text-embedding-3-small"
PINECONE_BM25_DEFAULT: Final[str] = "embedding:pinecone:bm25-default"


def _safe_register(name: str, register_fn: Callable[[], object]) -> None:
    try:
        resolve_embedding_method(name)
    except ValueError:
        register_fn()


def _register_defaults() -> None:
    _safe_register(
        OPENAI_TEXT_EMBEDDING_3_SMALL,
        lambda: register_langchain_embedding(
            OPENAI_TEXT_EMBEDDING_3_SMALL,
            lambda: OpenAIEmbeddings(
                model="text-embedding-3-small",
                dimensions=512,
            ),
        ),
    )
    _safe_register(
        PINECONE_BM25_DEFAULT,
        lambda: register_pinecone_bm25_embedding(PINECONE_BM25_DEFAULT),
    )


_register_defaults()


__all__ = [
    "OPENAI_TEXT_EMBEDDING_3_SMALL",
    "PINECONE_BM25_DEFAULT",
]



"""Tests for ChatKit server factory and configuration resolution."""

from __future__ import annotations
from collections.abc import Mapping
from types import SimpleNamespace
from typing import Any
from unittest.mock import MagicMock, patch
import pytest
from orcheo_backend.app.chatkit import server as chatkit_server_module
from orcheo_backend.app.chatkit.server import (
    _coerce_config_set,
    _resolve_chatkit_backend,
    _resolve_chatkit_pool_settings,
    _resolve_chatkit_postgres_dsn,
    create_chatkit_server,
)


class MockDynaconf(Mapping[str, Any]):
    def __init__(self, data: dict[str, Any]) -> None:
        self._data = data

    def __getitem__(self, key: str) -> Any:
        return self._data[key]

    def __iter__(self) -> Any:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def get(self, key: str, default: Any = None) -> Any:
        return self._data.get(key, default)


def test_resolve_chatkit_backend_inputs() -> None:
    # Test Mapping
    assert _resolve_chatkit_backend({"CHATKIT_BACKEND": "postgres"}) == "postgres"
    assert _resolve_chatkit_backend({"chatkit_backend": "postgres"}) == "postgres"

    # Test Object
    assert (
        _resolve_chatkit_backend(SimpleNamespace(CHATKIT_BACKEND="postgres"))
        == "postgres"
    )
    assert (
        _resolve_chatkit_backend(SimpleNamespace(chatkit_backend="postgres"))
        == "postgres"
    )


def test_resolve_chatkit_backend_invalid() -> None:
    with pytest.raises(ValueError, match="must be either 'sqlite' or 'postgres'"):
        _resolve_chatkit_backend({"CHATKIT_BACKEND": "invalid"})


def test_resolve_chatkit_postgres_dsn() -> None:
    # Test Mapping
    assert (
        _resolve_chatkit_postgres_dsn({"POSTGRES_DSN": "postgresql://test"})
        == "postgresql://test"
    )
    assert (
        _resolve_chatkit_postgres_dsn({"postgres_dsn": "postgresql://test"})
        == "postgresql://test"
    )

    # Test Object
    assert (
        _resolve_chatkit_postgres_dsn(SimpleNamespace(POSTGRES_DSN="postgresql://test"))
        == "postgresql://test"
    )
    assert (
        _resolve_chatkit_postgres_dsn(SimpleNamespace(postgres_dsn="postgresql://test"))
        == "postgresql://test"
    )

    # Test missing
    with pytest.raises(ValueError, match="must be set"):
        _resolve_chatkit_postgres_dsn({})


def test_resolve_chatkit_pool_settings() -> None:
    defaults = (1, 10, 30.0, 300.0)
    # Test defaults
    assert _resolve_chatkit_pool_settings({}) == defaults

    # Test Mapping
    settings = {
        "POSTGRES_POOL_MIN_SIZE": 2,
        "POSTGRES_POOL_MAX_SIZE": 20,
        "POSTGRES_POOL_TIMEOUT": 60.0,
        "POSTGRES_POOL_MAX_IDLE": 600.0,
    }
    assert _resolve_chatkit_pool_settings(settings) == (2, 20, 60.0, 600.0)

    # Test Object
    obj = SimpleNamespace(
        postgres_pool_min_size=2,
        postgres_pool_max_size=20,
        postgres_pool_timeout=60.0,
        postgres_pool_max_idle=600.0,
    )
    assert _resolve_chatkit_pool_settings(obj) == (2, 20, 60.0, 600.0)


def test_create_chatkit_server_postgres(monkeypatch: pytest.MonkeyPatch) -> None:
    mock_postgres_store = MagicMock()
    monkeypatch.setattr(
        chatkit_server_module, "PostgresChatKitStore", mock_postgres_store
    )

    repository = MagicMock()
    vault_provider = MagicMock()

    settings = {
        "CHATKIT_BACKEND": "postgres",
        "POSTGRES_DSN": "postgresql://test",
    }

    with patch("orcheo_backend.app.chatkit.server.get_settings", return_value=settings):
        create_chatkit_server(repository, vault_provider)

    mock_postgres_store.assert_called_once()
    args, kwargs = mock_postgres_store.call_args
    assert args[0] == "postgresql://test"


def test_coerce_config_set() -> None:
    default = {"default"}
    assert _coerce_config_set(None, default) == default
    assert _coerce_config_set("a, b, c", default) == {"a", "b", "c"}
    assert _coerce_config_set(["a", "b"], default) == {"a", "b"}
    assert _coerce_config_set([], default) == default
    assert _coerce_config_set(" ", default) == default


def test_resolve_backend_dynaconf() -> None:
    class FakeDynaconf:
        def get(self, key: str, default: Any = None) -> Any:
            return None

    # We patch the class in the module so isinstance checks work
    with patch("orcheo_backend.app.chatkit.server.Dynaconf", FakeDynaconf):
        settings = FakeDynaconf()
        # Override get for this instance
        object.__setattr__(
            settings,
            "get",
            lambda k, d=None: "postgres" if k == "CHATKIT_BACKEND" else None,
        )

        assert _resolve_chatkit_backend(settings) == "postgres"


def test_resolve_chatkit_postgres_dsn_dynaconf() -> None:
    class FakeDynaconf:
        def get(self, key: str, default: Any = None) -> Any:
            return "postgresql://test" if key == "POSTGRES_DSN" else None

    with patch("orcheo_backend.app.chatkit.server.Dynaconf", FakeDynaconf):
        settings = FakeDynaconf()
        assert _resolve_chatkit_postgres_dsn(settings) == "postgresql://test"


def test_resolve_chatkit_pool_settings_dynaconf() -> None:
    class FakeDynaconf:
        def get(self, key: str, default: Any = None) -> Any:
            mapping = {
                "POSTGRES_POOL_MIN_SIZE": 5,
                "POSTGRES_POOL_MAX_SIZE": 50,
                "POSTGRES_POOL_TIMEOUT": 10.0,
                "POSTGRES_POOL_MAX_IDLE": 100.0,
            }
            return mapping.get(key, default)

    with patch("orcheo_backend.app.chatkit.server.Dynaconf", FakeDynaconf):
        settings = FakeDynaconf()
        assert _resolve_chatkit_pool_settings(settings) == (5, 50, 10.0, 100.0)



"""Async-safe in-memory history store implementation."""

from __future__ import annotations
import asyncio
from collections.abc import Mapping
from datetime import datetime
from typing import Any
from orcheo_backend.app.history.models import (
    RunHistoryError,
    RunHistoryNotFoundError,
    RunHistoryRecord,
    RunHistoryStep,
)


class InMemoryRunHistoryStore:
    """Async-safe in-memory store for execution histories."""

    def __init__(self) -> None:
        """Initialize the in-memory store."""
        self._lock = asyncio.Lock()
        self._histories: dict[str, RunHistoryRecord] = {}

    async def start_run(
        self,
        *,
        workflow_id: str,
        execution_id: str,
        inputs: Mapping[str, Any] | None = None,
        runnable_config: Mapping[str, Any] | None = None,
        tags: list[str] | None = None,
        callbacks: list[Any] | None = None,
        metadata: Mapping[str, Any] | None = None,
        run_name: str | None = None,
        trace_id: str | None = None,
        trace_started_at: datetime | None = None,
    ) -> RunHistoryRecord:
        """Initialise a history record for the provided execution."""
        async with self._lock:
            if execution_id in self._histories:
                msg = f"History already exists for execution_id={execution_id}"
                raise RunHistoryError(msg)

            effective_trace_started = trace_started_at
            config_payload = runnable_config
            if runnable_config and hasattr(runnable_config, "model_dump"):
                config_payload = runnable_config.model_dump(mode="json")  # type: ignore[arg-type]
            tag_values = (
                list(tags or config_payload.get("tags", []))
                if isinstance(config_payload, Mapping)
                else list(tags or [])
            )
            callback_values = (
                list(callbacks or config_payload.get("callbacks", []))
                if isinstance(config_payload, Mapping)
                else list(callbacks or [])
            )
            metadata_values = (
                dict(metadata or config_payload.get("metadata", {}))
                if isinstance(config_payload, Mapping)
                else dict(metadata or {})
            )
            run_identifier = run_name or (
                config_payload.get("run_name")
                if isinstance(config_payload, Mapping)
                else None
            )
            record = RunHistoryRecord(
                workflow_id=workflow_id,
                execution_id=execution_id,
                inputs=dict(inputs or {}),
                runnable_config=dict(config_payload or {}),
                tags=tag_values,
                callbacks=callback_values,
                metadata=metadata_values,
                run_name=run_identifier,
                trace_id=trace_id,
                trace_started_at=effective_trace_started,
                trace_last_span_at=effective_trace_started,
            )
            if record.trace_started_at is None:
                record.trace_started_at = record.started_at
            if record.trace_last_span_at is None:
                record.trace_last_span_at = record.trace_started_at
            self._histories[execution_id] = record
            return record.model_copy(deep=True)

    async def append_step(
        self,
        execution_id: str,
        payload: Mapping[str, Any],
    ) -> RunHistoryStep:
        """Append a step for the execution."""
        async with self._lock:
            record = self._require_record(execution_id)
            return record.append_step(payload)

    async def mark_completed(self, execution_id: str) -> RunHistoryRecord:
        """Mark the execution as completed."""
        async with self._lock:
            record = self._require_record(execution_id)
            record.mark_completed()
            return record.model_copy(deep=True)

    async def mark_failed(self, execution_id: str, error: str) -> RunHistoryRecord:
        """Mark the execution as failed with the specified error message."""
        async with self._lock:
            record = self._require_record(execution_id)
            record.mark_failed(error)
            return record.model_copy(deep=True)

    async def mark_cancelled(
        self,
        execution_id: str,
        *,
        reason: str | None = None,
    ) -> RunHistoryRecord:
        """Mark the execution as cancelled."""
        async with self._lock:
            record = self._require_record(execution_id)
            record.mark_cancelled(reason=reason)
            return record.model_copy(deep=True)

    async def get_history(self, execution_id: str) -> RunHistoryRecord:
        """Return a deep copy of the execution history."""
        async with self._lock:
            record = self._require_record(execution_id)
            return record.model_copy(deep=True)

    async def clear(self) -> None:
        """Clear all stored histories. Intended for testing only."""
        async with self._lock:
            self._histories.clear()

    async def list_histories(
        self,
        workflow_id: str,
        *,
        limit: int | None = None,
    ) -> list[RunHistoryRecord]:
        """Return histories associated with the provided workflow."""
        async with self._lock:
            records = [
                record.model_copy(deep=True)
                for record in self._histories.values()
                if record.workflow_id == workflow_id
            ]

        records.sort(key=lambda record: record.started_at, reverse=True)
        if limit is not None:
            return records[:limit]
        return records

    def _require_record(self, execution_id: str) -> RunHistoryRecord:
        """Return the record or raise an error if missing."""
        record = self._histories.get(execution_id)
        if record is None:
            msg = f"History not found for execution_id={execution_id}"
            raise RunHistoryNotFoundError(msg)
        return record


__all__ = ["InMemoryRunHistoryStore"]



"""MCP tool implementations that wrap Orcheo service layer."""

from __future__ import annotations
from importlib import import_module as _import_module
from importlib import util as _util
from .agent_tools import (
    _ensure_agent_tools_registered,
    list_agent_tools,
    show_agent_tool,
)
from .agent_tools import (
    logger as _agent_tools_logger,
)
from .codegen import generate_workflow_scaffold, generate_workflow_template
from .credentials import create_credential, delete_credential, list_credentials
from .edges import list_edges, show_edge
from .nodes import list_nodes, show_node
from .service_tokens import (
    create_service_token,
    list_service_tokens,
    revoke_service_token,
    rotate_service_token,
    show_service_token,
)
from .workflow import (
    delete_workflow,
    download_workflow,
    list_workflows,
    publish_workflow,
    run_workflow,
    show_workflow,
    unpublish_workflow,
    upload_workflow,
)


import_module = _import_module
util = _util
logger = _agent_tools_logger

__all__ = [
    "_ensure_agent_tools_registered",
    "create_credential",
    "create_service_token",
    "delete_credential",
    "delete_workflow",
    "download_workflow",
    "generate_workflow_scaffold",
    "generate_workflow_template",
    "list_agent_tools",
    "list_credentials",
    "list_edges",
    "list_nodes",
    "list_service_tokens",
    "list_workflows",
    "import_module",
    "logger",
    "revoke_service_token",
    "publish_workflow",
    "rotate_service_token",
    "run_workflow",
    "show_agent_tool",
    "show_edge",
    "show_node",
    "show_service_token",
    "show_workflow",
    "unpublish_workflow",
    "upload_workflow",
    "util",
]



"""Tests for BaseNode fallback resolution logic."""

from __future__ import annotations
from typing import cast
from orcheo.graph.state import State
from orcheo.nodes.base import BaseNode


def test_fallback_to_results_requires_mapping() -> None:
    state = cast(State, {"results": ["unexpected"]})

    fallback = BaseNode._fallback_to_results(["output", "value"], 0, state)

    assert fallback is None



"""Schema creation and migrations for the SQLite ChatKit store."""

from __future__ import annotations
import json
import logging
import aiosqlite


SCHEMA_SQL = """
PRAGMA journal_mode = WAL;
CREATE TABLE IF NOT EXISTS chat_threads (
    id TEXT PRIMARY KEY,
    title TEXT,
    workflow_id TEXT,
    status_json TEXT NOT NULL,
    metadata_json TEXT NOT NULL,
    created_at TEXT NOT NULL,
    updated_at TEXT NOT NULL
);
CREATE TABLE IF NOT EXISTS chat_messages (
    id TEXT PRIMARY KEY,
    thread_id TEXT NOT NULL,
    ordinal INTEGER NOT NULL,
    item_type TEXT,
    item_json TEXT NOT NULL,
    created_at TEXT NOT NULL,
    FOREIGN KEY(thread_id) REFERENCES chat_threads(id)
        ON DELETE CASCADE
);
CREATE INDEX IF NOT EXISTS idx_chat_messages_thread
    ON chat_messages(thread_id, ordinal);
CREATE TABLE IF NOT EXISTS chat_attachments (
    id TEXT PRIMARY KEY,
    thread_id TEXT,
    attachment_type TEXT NOT NULL,
    name TEXT NOT NULL,
    mime_type TEXT NOT NULL,
    details_json TEXT NOT NULL,
    storage_path TEXT,
    created_at TEXT NOT NULL,
    FOREIGN KEY(thread_id) REFERENCES chat_threads(id)
        ON DELETE SET NULL
);
CREATE INDEX IF NOT EXISTS idx_chat_attachments_thread
    ON chat_attachments(thread_id);
"""

logger = logging.getLogger(__name__)


async def ensure_schema(conn: aiosqlite.Connection) -> None:
    """Ensure tables and indexes exist."""
    await run_migrations(conn)
    await conn.executescript(SCHEMA_SQL)
    await conn.commit()


async def run_migrations(conn: aiosqlite.Connection) -> None:
    """Apply pending migrations."""
    await _migrate_chat_messages_thread_id(conn)


async def _migrate_chat_messages_thread_id(conn: aiosqlite.Connection) -> None:
    cursor = await conn.execute(
        """
        SELECT name
          FROM sqlite_master
         WHERE type = 'table' AND name = 'chat_messages'
        """
    )
    table = await cursor.fetchone()
    if table is None:
        return

    cursor = await conn.execute("PRAGMA table_info(chat_messages)")
    columns = {row[1] for row in await cursor.fetchall()}
    if "thread_id" in columns:
        return

    logger.info("Migrating chat_messages table to add thread_id column")
    cursor = await conn.execute(
        """
        SELECT id, ordinal, item_type, item_json, created_at
          FROM chat_messages
         ORDER BY ordinal
        """
    )
    rows = await cursor.fetchall()

    await conn.execute("PRAGMA foreign_keys = OFF;")
    await conn.execute("DROP TABLE IF EXISTS chat_messages__new;")
    await conn.execute("BEGIN;")
    try:
        await conn.execute(
            """
            CREATE TABLE chat_messages__new (
                id TEXT PRIMARY KEY,
                thread_id TEXT NOT NULL,
                ordinal INTEGER NOT NULL,
                item_type TEXT,
                item_json TEXT NOT NULL,
                created_at TEXT NOT NULL,
                FOREIGN KEY(thread_id) REFERENCES chat_threads(id)
                    ON DELETE CASCADE
            )
            """
        )

        for row in rows:
            payload = json.loads(row["item_json"])
            thread_id = (
                payload.get("thread_id")
                or payload.get("threadId")
                or payload.get("metadata", {}).get("thread_id")
            )
            if thread_id is None:
                logger.warning(
                    (
                        "Dropping message %s while migrating chat_messages; "
                        "missing thread_id"
                    ),
                    row["id"],
                )
                continue

            await conn.execute(
                """
                INSERT INTO chat_messages__new (
                    id,
                    thread_id,
                    ordinal,
                    item_type,
                    item_json,
                    created_at
                ) VALUES (?, ?, ?, ?, ?, ?)
                """,
                (
                    row["id"],
                    str(thread_id),
                    row["ordinal"],
                    row["item_type"],
                    row["item_json"],
                    row["created_at"],
                ),
            )

        await conn.execute("DROP TABLE chat_messages;")
        await conn.execute("ALTER TABLE chat_messages__new RENAME TO chat_messages;")
        await conn.commit()
        logger.info("chat_messages migration completed successfully")
    except Exception:
        await conn.rollback()
        logger.exception(
            "Failed to migrate chat_messages table to include thread_id column"
        )
        raise
    finally:
        await conn.execute("PRAGMA foreign_keys = ON;")


__all__ = ["ensure_schema", "run_migrations"]



"""Webhook trigger state management and validation logic."""

from __future__ import annotations
import hashlib
import hmac
import json
from collections import deque
from collections.abc import Mapping
from datetime import UTC, datetime, timedelta
from typing import Any
from orcheo.triggers.webhook.config import WebhookTriggerConfig
from orcheo.triggers.webhook.errors import (
    MethodNotAllowedError,
    RateLimitExceededError,
    WebhookAuthenticationError,
    WebhookValidationError,
)
from orcheo.triggers.webhook.request import WebhookRequest


class WebhookTriggerState:
    """Maintain webhook configuration and request validation state."""

    def __init__(self, config: WebhookTriggerConfig | None = None) -> None:
        """Initialize state with an optional configuration instance."""
        self._config = (config or WebhookTriggerConfig()).model_copy(deep=True)
        self._recent_invocations: deque[datetime] = deque()
        self._recent_signatures: deque[tuple[str, datetime]] = deque()
        self._signature_cache: set[str] = set()

    @property
    def config(self) -> WebhookTriggerConfig:
        """Return a deep copy of the stored webhook configuration."""
        return self._config.model_copy(deep=True)

    def update_config(self, config: WebhookTriggerConfig) -> None:
        """Replace the configuration and reset rate limiting state."""
        self._config = config.model_copy(deep=True)
        self._recent_invocations.clear()
        self._recent_signatures.clear()
        self._signature_cache.clear()

    def validate(self, request: WebhookRequest) -> None:
        """Validate the inbound request against the configured rules."""
        self._validate_method(request)
        self._validate_required_headers(request)
        self._validate_required_query_params(request)
        self._validate_authentication(request)
        self._enforce_rate_limit()

    def serialize_payload(self, payload: Any) -> Any:
        """Normalize payloads for storage on workflow runs."""
        if isinstance(payload, bytes):
            decoded = payload.decode("utf-8", errors="replace")
            return {"raw": decoded}
        return payload

    def scrub_headers_for_storage(self, headers: Mapping[str, str]) -> dict[str, str]:
        """Redact sensitive headers such as shared secrets before storage."""
        sanitized = {key: value for key, value in headers.items()}
        secret_header = self._config.shared_secret_header
        if secret_header:
            sanitized.pop(secret_header, None)
        signature_header = self._config.hmac_header
        if signature_header:
            sanitized.pop(signature_header, None)
        return sanitized

    # Internal helpers -------------------------------------------------

    def _validate_method(self, request: WebhookRequest) -> None:
        allowed = set(self._config.allowed_methods)
        method = request.normalized_method()
        if method not in allowed:
            raise MethodNotAllowedError(method, allowed)

    def _validate_authentication(self, request: WebhookRequest) -> None:
        if self._config.hmac_secret:
            self._validate_hmac_signature(request)
        if self._config.shared_secret:
            self._validate_shared_secret(request)

    def _validate_shared_secret(self, request: WebhookRequest) -> None:
        header_name = self._config.shared_secret_header
        if header_name is None:
            return  # pragma: no cover - defensive

        expected = self._config.shared_secret
        provided = request.normalized_headers().get(header_name)
        if expected is None or provided is None:
            raise WebhookAuthenticationError()
        if not hmac.compare_digest(provided, expected):
            raise WebhookAuthenticationError()

    def _validate_hmac_signature(self, request: WebhookRequest) -> None:
        header_name = self._config.hmac_header
        secret = self._config.hmac_secret
        if header_name is None or not secret:
            return

        headers = request.normalized_headers()
        provided_raw = headers.get(header_name)
        if not provided_raw:
            raise WebhookAuthenticationError()
        signature = self._extract_hmac_signature(provided_raw)

        timestamp_header = self._config.hmac_timestamp_header
        if timestamp_header:
            timestamp_value = headers.get(timestamp_header)
            if not timestamp_value:
                raise WebhookAuthenticationError()
            timestamp = self._parse_signature_timestamp(timestamp_value)
        else:
            timestamp = None

        payload_bytes = self._canonical_payload_bytes(request.payload)
        components: list[bytes] = []
        if timestamp is not None:
            components.append(str(int(timestamp.timestamp())).encode("utf-8"))
        components.append(payload_bytes)
        message = b".".join(components)

        hasher = getattr(hashlib, self._config.hmac_algorithm)
        expected = hmac.new(secret.encode("utf-8"), message, hasher).hexdigest()
        if not hmac.compare_digest(expected, signature):
            raise WebhookAuthenticationError()

        self._enforce_signature_replay(signature, timestamp)

    def _canonical_payload_bytes(self, payload: Any) -> bytes:
        if payload is None:
            return b""
        if isinstance(payload, bytes):
            return payload
        if isinstance(payload, str):
            return payload.encode("utf-8")
        try:
            serialized = json.dumps(payload, sort_keys=True, separators=(",", ":"))
        except (TypeError, ValueError):  # pragma: no cover - defensive
            serialized = str(payload)
        return serialized.encode("utf-8")

    def _extract_hmac_signature(self, raw: str) -> str:
        candidate = raw.strip()
        if not candidate:
            raise WebhookAuthenticationError()
        for segment in candidate.split(","):
            part = segment.strip()
            if "=" in part:
                key, value = part.split("=", 1)
                if key.lower() in {"sha256", "sha512", "v1", "signature"}:
                    candidate = value.strip()
        if not candidate:
            raise WebhookAuthenticationError()
        return candidate

    def _parse_signature_timestamp(self, value: str) -> datetime:
        stripped = value.strip()
        if not stripped:
            raise WebhookAuthenticationError()
        if stripped.isdigit():
            timestamp = datetime.fromtimestamp(int(stripped), tz=UTC)
        else:
            try:
                timestamp = datetime.fromisoformat(stripped.replace("Z", "+00:00"))
            except ValueError as exc:  # pragma: no cover - defensive
                raise WebhookAuthenticationError() from exc

        tolerance = self._config.hmac_tolerance_seconds
        if tolerance:
            now = datetime.now(tz=UTC)
            if abs((now - timestamp).total_seconds()) > tolerance:
                raise WebhookAuthenticationError()
        return timestamp

    def _enforce_signature_replay(
        self, signature: str, timestamp: datetime | None
    ) -> None:
        tolerance = max(self._config.hmac_tolerance_seconds, 1)
        now = datetime.now(tz=UTC)
        cutoff = now - timedelta(seconds=tolerance)
        while self._recent_signatures and self._recent_signatures[0][1] < cutoff:
            old_signature, _ = self._recent_signatures.popleft()
            self._signature_cache.discard(old_signature)
        if signature in self._signature_cache:
            raise WebhookAuthenticationError()
        self._signature_cache.add(signature)
        self._recent_signatures.append((signature, timestamp or now))

    def _validate_required_headers(self, request: WebhookRequest) -> None:
        expected = self._config.required_headers
        if not expected:
            return

        headers = request.normalized_headers()
        for key, value in expected.items():
            if headers.get(key) != value:
                message = f"Missing or invalid required header: {key}"
                raise WebhookValidationError(message, status_code=400)

    def _validate_required_query_params(self, request: WebhookRequest) -> None:
        expected = self._config.required_query_params
        if not expected:
            return

        params = request.normalized_query()
        for key, value in expected.items():
            if params.get(key) != value:
                message = f"Missing or invalid required query parameter: {key}"
                raise WebhookValidationError(message, status_code=400)

    def _enforce_rate_limit(self) -> None:
        config = self._config.rate_limit
        if config is None:
            return

        now = datetime.now(tz=UTC)
        window_start = now - timedelta(seconds=config.interval_seconds)

        while self._recent_invocations and self._recent_invocations[0] < window_start:
            self._recent_invocations.popleft()

        if len(self._recent_invocations) >= config.limit:
            raise RateLimitExceededError(config.limit, config.interval_seconds)

        self._recent_invocations.append(now)


__all__ = ["WebhookTriggerState"]



"""Workflow upload CLI tests for LangGraph error scenarios."""

from __future__ import annotations
import json
from pathlib import Path
import httpx
import respx
from typer.testing import CliRunner
from orcheo_sdk.cli.errors import CLIError
from orcheo_sdk.cli.main import app


def test_workflow_upload_langgraph_script_fetch_existing_error(
    runner: CliRunner, env: dict[str, str], tmp_path: Path
) -> None:
    """Test workflow upload fails when fetching existing workflow fails."""
    py_file = tmp_path / "langgraph_workflow.py"
    py_file.write_text(
        """
from langgraph.graph import StateGraph

def greet(state):
    return {"message": "Hello"}
""",
        encoding="utf-8",
    )

    with respx.mock(assert_all_called=True) as router:
        router.get("http://api.test/api/workflows/wf-1").mock(
            return_value=httpx.Response(404, json={"error": "Not found"})
        )
        result = runner.invoke(
            app,
            ["workflow", "upload", str(py_file), "--id", "wf-1"],
            env=env,
        )
    assert result.exit_code != 0
    assert isinstance(result.exception, CLIError)
    assert "Failed to fetch workflow" in str(result.exception)


def test_workflow_upload_langgraph_script_create_workflow_error(
    runner: CliRunner, env: dict[str, str], tmp_path: Path
) -> None:
    """Test workflow upload fails when creating new workflow fails."""
    py_file = tmp_path / "langgraph_workflow.py"
    py_file.write_text(
        """
from langgraph.graph import StateGraph

def greet(state):
    return {"message": "Hello"}
""",
        encoding="utf-8",
    )

    with respx.mock(assert_all_called=True) as router:
        router.post("http://api.test/api/workflows").mock(
            return_value=httpx.Response(500, json={"error": "Internal error"})
        )
        result = runner.invoke(
            app,
            ["workflow", "upload", str(py_file)],
            env=env,
        )
    assert result.exit_code != 0
    assert isinstance(result.exception, CLIError)
    assert "Failed to create workflow" in str(result.exception)


def test_workflow_upload_with_blank_name_error(
    runner: CliRunner, env: dict[str, str], tmp_path: Path
) -> None:
    """Test workflow upload rejects empty rename values."""
    json_file = tmp_path / "workflow.json"
    json_file.write_text(
        json.dumps({"name": "Original", "graph": {"nodes": [], "edges": []}}),
        encoding="utf-8",
    )

    result = runner.invoke(
        app,
        ["workflow", "upload", str(json_file), "--name", "   "],
        env=env,
    )
    assert result.exit_code != 0
    assert isinstance(result.exception, CLIError)
    assert "cannot be empty" in str(result.exception)



"""Storage nodes providing database access."""

from __future__ import annotations
import asyncio
import sqlite3
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, Literal, cast
from langchain_core.runnables import RunnableConfig
from pydantic import Field
from orcheo.graph.state import State
from orcheo.nodes.base import TaskNode
from orcheo.nodes.registry import NodeMetadata, registry


if TYPE_CHECKING:
    pass


def _rows_to_dicts(
    columns: Sequence[str], rows: Sequence[Sequence[Any]]
) -> list[dict[str, Any]]:
    """Convert database rows into dictionaries keyed by column name."""
    return [dict(zip(columns, row, strict=False)) for row in rows]


@registry.register(
    NodeMetadata(
        name="PostgresNode",
        description="Execute SQL against a PostgreSQL database using psycopg.",
        category="storage",
    )
)
class PostgresNode(TaskNode):
    """Node encapsulating basic PostgreSQL interactions."""

    dsn: str = Field(
        default="[[postgres_dsn]]",
        description="PostgreSQL DSN, e.g. postgresql://user:pass@host/db",
    )
    query: str = Field(description="SQL query to execute")
    parameters: Mapping[str, Any] | Sequence[Any] | None = Field(
        default=None, description="Parameters bound to the SQL query"
    )
    fetch: Literal["none", "one", "all"] = Field(
        default="all", description="Fetch strategy for returning result rows"
    )
    autocommit: bool = Field(
        default=True, description="Enable autocommit mode for the connection"
    )

    def _execute(self) -> dict[str, Any]:
        """Execute the configured query returning structured results."""
        import psycopg

        with psycopg.connect(self.dsn) as connection:
            connection.autocommit = self.autocommit
            with connection.cursor() as cursor:
                cursor.execute(self.query, self.parameters)

                if self.fetch == "none":
                    return {"rows": [], "rowcount": cursor.rowcount}

                row = cursor.fetchone() if self.fetch == "one" else cursor.fetchall()
                if cursor.description is None:
                    if self.fetch == "one":
                        if row is None:
                            raw_rows: list[Sequence[Any]] = []
                        else:
                            raw_rows = [cast(Sequence[Any], row)]
                    else:
                        raw_rows = list(cast(Sequence[Sequence[Any]], row))
                    return {"rows": raw_rows, "rowcount": cursor.rowcount}

                columns = [column.name for column in cursor.description]
                if self.fetch == "one":
                    if row is None:
                        data_rows: list[Sequence[Any]] = []
                    else:
                        data_rows = [cast(Sequence[Any], row)]
                else:
                    data_rows = list(cast(Sequence[Sequence[Any]], row))

                if data_rows:
                    first_row = data_rows[0]
                    if len(columns) != len(first_row):
                        return {"rows": data_rows, "rowcount": cursor.rowcount}

                mapped_rows = _rows_to_dicts(columns, data_rows)
                return {"rows": mapped_rows, "rowcount": cursor.rowcount}

    async def run(self, state: State, config: RunnableConfig) -> dict[str, Any]:
        """Execute the SQL query asynchronously."""
        return await asyncio.to_thread(self._execute)


@registry.register(
    NodeMetadata(
        name="SQLiteNode",
        description="Execute SQL statements against a SQLite database.",
        category="storage",
    )
)
class SQLiteNode(TaskNode):
    """Node providing simple SQLite access suitable for local workflows."""

    database: str = Field(default=":memory:", description="SQLite database path")
    query: str = Field(description="SQL query to execute")
    parameters: Mapping[str, Any] | Sequence[Any] | None = Field(
        default=None, description="Parameters bound to the SQL query"
    )
    fetch: Literal["none", "one", "all"] = Field(
        default="all", description="Fetch strategy for returning result rows"
    )

    def _execute(self) -> dict[str, Any]:
        """Execute the SQL query returning structured results."""
        connection = sqlite3.connect(self.database)
        connection.row_factory = sqlite3.Row
        try:
            cursor = connection.execute(self.query, self.parameters or [])
            connection.commit()

            if self.fetch == "none":
                return {"rows": [], "rowcount": cursor.rowcount}

            if self.fetch == "one":
                row = cursor.fetchone()
                if row is None:
                    return {"rows": [], "rowcount": cursor.rowcount}
                return {
                    "rows": [dict(row)],
                    "rowcount": cursor.rowcount,
                }

            rows = cursor.fetchall()
            return {
                "rows": [dict(item) for item in rows],
                "rowcount": cursor.rowcount,
            }
        finally:
            connection.close()

    async def run(self, state: State, config: RunnableConfig) -> dict[str, Any]:
        """Execute the SQLite query asynchronously."""
        return await asyncio.to_thread(self._execute)


__all__ = ["PostgresNode", "SQLiteNode"]



"""Base edge implementation for Orcheo."""

from abc import abstractmethod
from langchain_core.runnables import RunnableConfig
from langgraph.types import Send
from orcheo.graph.state import State
from orcheo.nodes.base import BaseRunnable


class BaseEdge(BaseRunnable):
    """Base class for all edges in the flow.

    Edges are responsible for routing decisions in the workflow graph.
    They inherit variable decoding and credential resolution from BaseRunnable
    but return routing decisions instead of data transformations.
    """

    async def __call__(self, state: State, config: RunnableConfig) -> str | list[Send]:
        """Execute the edge and return the routing decision."""
        self.decode_variables(state, config=config)
        result = await self.run(state, config)
        return result

    @abstractmethod
    async def run(self, state: State, config: RunnableConfig) -> str | list[Send]:
        """Run the edge and return routing decision.

        Returns:
            Either a string indicating the next node/edge to route to,
            or a list of Send objects for dynamic fan-out patterns.
        """
        pass  # pragma: no cover


__all__ = ["BaseEdge"]



"""Tests for the workflow version ingestion endpoints."""

import asyncio
import textwrap
from uuid import UUID, uuid4
import pytest
from fastapi import HTTPException, status
from fastapi.testclient import TestClient
from orcheo.graph.ingestion import LANGGRAPH_SCRIPT_FORMAT
from orcheo_backend.app import create_app, ingest_workflow_version
from orcheo_backend.app.authentication import reset_authentication_state
from orcheo_backend.app.repository import (
    InMemoryWorkflowRepository,
    WorkflowNotFoundError,
)
from orcheo_backend.app.schemas.workflows import WorkflowVersionIngestRequest


def test_ingest_workflow_version_endpoint_creates_version(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """LangGraph scripts can be submitted to create workflow versions."""

    monkeypatch.setenv("ORCHEO_AUTH_MODE", "disabled")
    reset_authentication_state()

    repository = InMemoryWorkflowRepository()
    workflow = asyncio.run(
        repository.create_workflow(
            name="LangGraph", slug=None, description=None, tags=[], actor="tester"
        )
    )

    app = create_app(repository)
    client = TestClient(app)

    script = textwrap.dedent(
        """
        from langgraph.graph import StateGraph
        from orcheo.graph.state import State

        def build_graph():
            graph = StateGraph(State)
            graph.add_node("noop", lambda state: state)
            graph.set_entry_point("noop")
            graph.set_finish_point("noop")
            return graph
        """
    )

    response = client.post(
        f"/api/workflows/{workflow.id}/versions/ingest",
        json={
            "script": script,
            "entrypoint": "build_graph",
            "metadata": {"language": "python"},
            "notes": "Initial LangGraph import",
            "created_by": "tester",
        },
    )

    assert response.status_code == 201
    version = response.json()
    assert version["metadata"] == {"language": "python"}
    assert version["notes"] == "Initial LangGraph import"
    assert version["graph"]["format"] == LANGGRAPH_SCRIPT_FORMAT
    assert "summary" in version["graph"]


def test_ingest_workflow_version_invalid_script_returns_400(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Invalid LangGraph scripts return a 400 error."""

    monkeypatch.setenv("ORCHEO_AUTH_MODE", "disabled")
    reset_authentication_state()

    repository = InMemoryWorkflowRepository()
    workflow = asyncio.run(
        repository.create_workflow(
            name="Bad Script", slug=None, description=None, tags=[], actor="tester"
        )
    )

    app = create_app(repository)
    client = TestClient(app)

    invalid_script = "this is not valid python code!!!"

    response = client.post(
        f"/api/workflows/{workflow.id}/versions/ingest",
        json={
            "script": invalid_script,
            "entrypoint": "build_graph",
            "created_by": "tester",
        },
    )

    assert response.status_code == 400
    assert "detail" in response.json()


def test_ingest_workflow_version_missing_workflow_returns_404(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Ingesting a script for a non-existent workflow returns 404."""

    monkeypatch.setenv("ORCHEO_AUTH_MODE", "disabled")
    reset_authentication_state()

    repository = InMemoryWorkflowRepository()
    app = create_app(repository)
    client = TestClient(app)

    missing_id = str(uuid4())

    script = textwrap.dedent(
        """
        from langgraph.graph import StateGraph
        from orcheo.graph.state import State

        def build_graph():
            graph = StateGraph(State)
            graph.add_node("noop", lambda state: state)
            graph.set_entry_point("noop")
            graph.set_finish_point("noop")
            return graph
        """
    )

    response = client.post(
        f"/api/workflows/{missing_id}/versions/ingest",
        json={
            "script": script,
            "entrypoint": "build_graph",
            "created_by": "tester",
        },
    )

    assert response.status_code == 404
    assert response.json()["detail"] == "Workflow not found"


@pytest.mark.asyncio
async def test_ingest_workflow_version_raises_not_found_error() -> None:
    """Repository lookups raising ``WorkflowNotFoundError`` propagate as 404s."""

    script = textwrap.dedent(
        """
        from langgraph.graph import StateGraph
        from orcheo.graph.state import State

        def build_graph():
            graph = StateGraph(State)
            graph.add_node("noop", lambda state: state)
            graph.set_entry_point("noop")
            graph.set_finish_point("noop")
            return graph
        """
    )

    request = WorkflowVersionIngestRequest(
        script=script,
        entrypoint="build_graph",
        created_by="tester",
    )

    class FailingRepository(InMemoryWorkflowRepository):
        async def create_version(
            self,
            workflow_id: UUID,
            *,
            graph: dict[str, object],
            metadata: dict[str, object],
            runnable_config: dict[str, object] | None = None,
            notes: str | None,
            created_by: str,
        ):
            raise WorkflowNotFoundError(str(workflow_id))

    repository = FailingRepository()

    with pytest.raises(HTTPException) as exc_info:
        await ingest_workflow_version(uuid4(), request, repository)

    assert exc_info.value.status_code == status.HTTP_404_NOT_FOUND
    assert exc_info.value.detail == "Workflow not found"



"""Tests for the Agentensor checkpoint routers."""

from __future__ import annotations
from uuid import uuid4
import pytest
from fastapi import HTTPException
from orcheo.agentensor.checkpoints import (
    AgentensorCheckpoint,
    AgentensorCheckpointNotFoundError,
)
from orcheo_backend.app.routers.agentensor import (
    get_agentensor_checkpoint,
    list_agentensor_checkpoints,
)


class DummyStore:
    def __init__(self, checkpoint: AgentensorCheckpoint) -> None:
        self.checkpoint = checkpoint

    async def list_checkpoints(
        self, workflow_id: str, *, limit: int
    ) -> list[AgentensorCheckpoint]:
        return [self.checkpoint]

    async def get_checkpoint(self, checkpoint_id: str) -> AgentensorCheckpoint:
        if self.checkpoint.id != checkpoint_id:
            raise AgentensorCheckpointNotFoundError("missing")
        return self.checkpoint


def _build_checkpoint(workflow_id: str) -> AgentensorCheckpoint:
    return AgentensorCheckpoint(
        workflow_id=workflow_id,
        config_version=1,
        runnable_config={"alpha": "beta"},
        metrics={"score": 1.0},
        metadata={"stage": "test"},
        artifact_url="s3://bucket/checkpoint",
        is_best=True,
    )


@pytest.mark.asyncio
async def test_list_agentensor_checkpoints_returns_payloads() -> None:
    workflow_uuid = uuid4()
    checkpoint = _build_checkpoint(str(workflow_uuid))
    store = DummyStore(checkpoint)

    response = await list_agentensor_checkpoints(
        workflow_id=workflow_uuid,
        store=store,  # type: ignore[arg-type]
        limit=1,
    )

    assert len(response) == 1
    assert response[0].id == checkpoint.id
    assert response[0].workflow_id == str(workflow_uuid)


@pytest.mark.asyncio
async def test_get_agentensor_checkpoint_raises_not_found() -> None:
    workflow_id = uuid4()
    checkpoint = _build_checkpoint(str(workflow_id))
    store = DummyStore(checkpoint)

    with pytest.raises(HTTPException) as exc_info:
        await get_agentensor_checkpoint(
            workflow_id=uuid4(),  # mismatch on purpose
            checkpoint_id=uuid4().hex,
            store=store,  # type: ignore[arg-type]
        )

    assert exc_info.value.status_code == 404


@pytest.mark.asyncio
async def test_get_agentensor_checkpoint_requires_matching_workflow() -> None:
    workflow_id = uuid4()
    checkpoint = _build_checkpoint(str(workflow_id))
    store = DummyStore(checkpoint)

    with pytest.raises(HTTPException) as exc_info:
        await get_agentensor_checkpoint(
            workflow_id=uuid4(),
            checkpoint_id=checkpoint.id,
            store=store,  # type: ignore[arg-type]
        )

    assert exc_info.value.status_code == 404


@pytest.mark.asyncio
async def test_get_agentensor_checkpoint_returns_record() -> None:
    workflow_id = uuid4()
    checkpoint = _build_checkpoint(str(workflow_id))
    store = DummyStore(checkpoint)

    response = await get_agentensor_checkpoint(
        workflow_id=workflow_id,
        checkpoint_id=checkpoint.id,
        store=store,  # type: ignore[arg-type]
    )

    assert response.id == checkpoint.id
    assert response.workflow_id == str(workflow_id)
